import json
import urllib
import urllib.request
from dataclasses import dataclass

API_URL = "https://explain.dalibo.com/new.json"


@dataclass
class PevResponse:
    id: str
    delete_key: str

    @property
    def url(self) -> str:
        return f"https://explain.dalibo.com/plan/{self.id}"

    @property
    def delete_url(self) -> str:
        return f"{self.url}/{self.delete_key}"

    def delete(self) -> None:
        """Deletes the plan from explain.dalibo"""
        urllib.request.urlopen(urllib.request.Request(self.delete_url))

    def __repr__(self) -> str:
        return f"PevResult(url={self.url})"


def upload_sql_plan(query: str, plan: str, title: str) -> PevResponse:
    """Uploads a sql plan to explain.dalibo for visualization"""
    payload = json.dumps({"title": title, "plan": json.dumps(plan), "query": query}).encode()
    with urllib.request.urlopen(
        urllib.request.Request(
            API_URL,
            headers={
                "Content-Type": "application/json",
            },
            data=payload,
        )
    ) as response:
        data = json.loads(response.read())
        return PevResponse(id=data["id"], delete_key=data["deleteKey"])



# Generated by Django 4.2.9 on 2024-01-17 01:46
import logging

from django.db import OperationalError, connection, migrations


def create_extension(apps, schema_editor):
    try:
        with connection.cursor() as cursor:
            cursor.execute("CREATE EXTENSION IF NOT EXISTS pg_stat_statements")
    except OperationalError:
        logging.error("Could not install extension `pg_stat_statements`.")


class Migration(migrations.Migration):
    dependencies = []

    operations = [
        migrations.RunPython(create_extension, migrations.RunPython.noop),
    ]



# mypy: ignore-errors
from .utils import explain  # NOQA

try:
    import importlib.metadata as importlib_metadata  # type: ignore[import]
except ModuleNotFoundError:
    import importlib_metadata  # type: ignore[import]

__version__ = importlib_metadata.version(__name__)



import dataclasses
from datetime import datetime

from django.db import connection


@dataclasses.dataclass
class ConnectionInfo:
    pid: int
    database: str
    user: str
    source: str
    ip: str
    backend_start: datetime
    query_start: datetime | None
    wait_event: str
    state: str
    query: str
    ssl: bool


def get_connections_current_database() -> list[ConnectionInfo]:
    """Returns list of postgres connections"""
    sql = """
SELECT
    pg_stat_activity.pid,
    datname AS database,
    usename AS user,
    application_name AS source,
    client_addr AS ip,
    backend_start,
    query_start,
    wait_event,
    state,
    query,
    ssl
FROM
    pg_stat_activity
LEFT JOIN
    pg_stat_ssl ON pg_stat_activity.pid = pg_stat_ssl.pid
WHERE
    datname = %s
ORDER BY
    pg_stat_activity.pid
    """
    with connection.cursor() as cursor:
        database_name = connection.get_connection_params()["database"]
        cursor.execute(sql, [database_name])
        return list(ConnectionInfo(*c) for c in cursor.fetchall())



import dataclasses
from decimal import Decimal
from itertools import groupby
from pathlib import Path

from django.db import connection


@dataclasses.dataclass
class IndexInfo:
    schema: str
    table: str
    name: str
    columns: list[str]
    using: str
    unique: bool
    primary: bool
    valid: bool
    indexprs: str | None
    indpred: str | None
    definition: str
    hit_rate: Decimal
    index_scans: int
    size_bytes: int
    covered_by: "IndexInfo | None"

    @property
    def is_unused(self) -> bool:
        return self.index_scans < 50

    @property
    def is_duplicated(self) -> bool:
        return bool(self.covered_by)

    @property
    def columns_formatted(self) -> str:
        return ", ".join(self.columns)

    @property
    def table_key(self) -> str:
        return f"{self.schema}.{self.table}"

    def index_covers(self, columns: list[str]) -> bool:
        """Returns true if this index covers the given columns"""
        return self.columns[: len(columns)] == columns


def get_indexes() -> list[IndexInfo]:
    with connection.cursor() as cursor:
        sql = (Path(__file__).parent / "index_lookup.sql").read_text()
        cursor.execute(sql)
        indexes = list(IndexInfo(*c) for c in cursor.fetchall())
    _update_indexes_with_duplicated_indexes(indexes)
    return indexes


def get_total_index_hitrate() -> float:
    """Total hit rate with all indexes"""
    with connection.cursor() as cursor:
        cursor.execute(
            """
            SELECT
                (sum(idx_blks_hit)) / nullif(sum(idx_blks_hit + idx_blks_read), 0) AS rate
            FROM
                pg_statio_user_indexes
        """
        )
        return cursor.fetchone()[0]


def _update_indexes_with_duplicated_indexes(
    indexes: list[IndexInfo] | None = None,
) -> None:
    """Modifies a list of indexes a reference to the other index that covers it"""
    indexes = indexes or get_indexes()
    indexes_by_table = {key: list(group) for key, group in groupby(indexes, lambda i: i.table_key)}
    for index in indexes:
        if index.valid and not index.primary and not index.unique:
            for other_index in indexes_by_table[index.table_key]:
                if (
                    other_index.valid
                    and other_index.name != index.name
                    and other_index.index_covers(index.columns)
                    and other_index.using == index.using
                    and other_index.indexprs == index.indexprs
                    and other_index.indpred == index.indpred
                ):
                    index.covered_by = other_index


def get_index_stats(indexes: list[IndexInfo] | None = None) -> dict:
    indexes = indexes or get_indexes()

    ret = {
        "total_index_hitrate": get_total_index_hitrate() * 100,
        "total_indexes_size": sum(i.size_bytes for i in indexes),
        "total_duplicated_indexes_size": sum(i.size_bytes for i in indexes if i.is_duplicated),
        "total_unused_indexes_size": sum(i.size_bytes for i in indexes if i.is_unused),
        "count_indexes": len(indexes),
        "count_duplicated_indexes": sum(1 for i in indexes if i.is_duplicated),
        "count_unused_indexes": sum(1 for i in indexes if i.is_unused),
    }

    ret["percent_duplicated_indexes"] = float(ret["total_duplicated_indexes_size"]) / ret["total_indexes_size"] * 100  # type:ignore
    ret["percent_unused_indexes"] = float(ret["total_unused_indexes_size"]) / ret["total_indexes_size"] * 100  # type:ignore

    return ret



import dataclasses
from datetime import datetime

from django.db import connection


def get_tables_missing_indexes() -> list[tuple[str, str, str, int]]:
    with connection.cursor() as cursor:
        sql = """
    SELECT
        schemaname AS schema,
        relname AS table,
    FROM
        pg_stat_user_tables
    WHERE
        idx_scan > 0
        AND (100 * idx_scan / (seq_scan + idx_scan)) < 95
        AND n_live_tup >= 10000
    ORDER BY
        n_live_tup DESC,
        relname ASC
        """
        cursor.execute(sql)
        return cursor.fetchall()


@dataclasses.dataclass
class TableInfo:
    schema: str
    name: str
    last_vacuum: datetime | None
    last_autovacuum: datetime | None
    last_analyze: datetime | None
    last_autoanalyze: datetime | None
    dead_rows: int
    live_rows: int
    idx_scan: int
    seq_scan: int
    size_bytes: int
    row_estimate: int

    @property
    def requires_index(self) -> bool:
        return (
            self.idx_scan > 0
            and (100 * self.idx_scan / (self.seq_scan + self.idx_scan)) < 95
            and self.live_rows >= 10000
        )

    @property
    def index_hit_rate(self) -> float:
        return self.idx_scan / (self.seq_scan + self.idx_scan) if self.idx_scan > 0 else 0

    @property
    def index_hit_rate_formatted(self) -> str:
        return f"{self.index_hit_rate * 100: 0.2f}%" if self.index_hit_rate else ""


def get_maintenance_info() -> list[TableInfo]:
    sql = """
WITH space_stats AS (
    SELECT
        n.nspname AS schema,
        c.relname AS name,
        CASE WHEN c.relkind = 'r' THEN 'table' ELSE 'index' END AS type,
        pg_table_size(c.oid) AS size_bytes,
        (case WHEN reltuples < 0 THEN 0 else reltuples END)::int as row_estimate
    FROM
        pg_class c
    LEFT JOIN
        pg_namespace n ON n.oid = c.relnamespace
    WHERE
        n.nspname NOT IN ('pg_catalog', 'information_schema')
        AND n.nspname !~ '^pg_toast'
        AND c.relkind = 'r'
), maintenance_stats as (
    SELECT
        schemaname AS schema,
        relname AS name,
        last_vacuum,
        last_autovacuum,
        last_analyze,
        last_autoanalyze,
        n_dead_tup AS dead_rows,
        n_live_tup AS live_rows,
        idx_scan,
        seq_scan
    FROM
        pg_stat_user_tables
)
SELECT
    maintenance_stats.*,
    size_bytes,
    row_estimate
FROM
    maintenance_stats
JOIN
    space_stats
ON
    maintenance_stats.schema = space_stats.schema
    AND maintenance_stats.name = space_stats.name
"""
    with connection.cursor() as cursor:
        cursor.execute(sql)
        return [TableInfo(*row) for row in cursor.fetchall()]



import uuid
from typing import Any

from django import forms
from django.conf import settings
from django.contrib.auth.mixins import UserPassesTestMixin
from django.core.cache import cache
from django.http import HttpRequest, HttpResponseRedirect
from django.http.response import HttpResponse as HttpResponse
from django.shortcuts import render
from django.test import Client as TestClient
from django.urls import reverse
from django.views.generic import FormView, TemplateView

from .utils import ExplainSet, explain, indexes, live_connections, maintenance, queries, space

# Create your views here.


def get_cache_key(explainset_id: Any) -> str:
    return f"DJANGO_PEV:EXPLAIN:{explainset_id}"


def index(request):
    return render(request, "django_pev/index.html")


class BaseView(UserPassesTestMixin, TemplateView):
    template_name = "django_pev/indexes_view.html"

    def test_func(self):
        return self.request.user.is_superuser


class SpaceView(BaseView):
    template_name = "django_pev/space.html"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        ctx = super().get_context_data(**kwargs)

        ctx["database_size"] = space.get_database_size()
        ctx["tables"] = space.get_table_sizes()
        ctx["indexes"] = space.get_index_sizes()

        ctx["indexes_size"] = sum(c.size_bytes for c in ctx["indexes"])
        ctx["table_size"] = sum(c.size_bytes for c in ctx["tables"])

        return ctx


class MaintenanceView(BaseView):
    template_name = "django_pev/maintenance.html"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        ctx = super().get_context_data(**kwargs)

        ctx["tables"] = maintenance.get_maintenance_info()
        return ctx


class IndexesView(BaseView):
    template_name = "django_pev/indexes_view.html"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        ctx = super().get_context_data(**kwargs)
        all_indexes = indexes.get_indexes()
        ctx["all_indexes"] = all_indexes
        ctx["unused_indexes"] = sorted(
            [i for i in all_indexes if i.is_unused],
            key=lambda i: i.size_bytes,
            reverse=True,
        )
        ctx["duplicated_indexes"] = sorted(
            [i for i in all_indexes if i.is_duplicated],
            key=lambda i: i.size_bytes,
            reverse=True,
        )

        ctx.update(indexes.get_index_stats())
        return ctx


class ConnectionsView(BaseView):
    template_name = "django_pev/connections.html"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        ctx = super().get_context_data(**kwargs)
        ctx["connections"] = live_connections.get_connections_current_database()
        return ctx


class LiveQueriesView(BaseView):
    template_name = "django_pev/live_queries.html"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        ctx = super().get_context_data(**kwargs)
        ctx["queries"] = queries.get_live_queries_current_database()
        return ctx


class QueriesView(BaseView):
    template_name = "django_pev/queries.html"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        ctx = super().get_context_data(**kwargs)
        ctx["queries"] = queries.get_query_stats()
        ctx["is_pg_stats_enabled"] = queries.is_pg_stat_statements_installed()
        ctx["queries_by_io"] = sorted(queries.get_query_stats(), key=lambda i: i.shared_blks_hit, reverse=True)
        ctx["queries_by_slowest"] = sorted(queries.get_query_stats(), key=lambda i: i.mean_time, reverse=True)
        return ctx


class ExplainView(BaseView):
    template_name = "django_pev/explain.html"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        ctx = super().get_context_data(**kwargs)
        explain_result = None

        if url := self.request.GET.get("url"):
            client = TestClient()
            client.force_login(self.request.user)  # type:ignore

            with explain(url=url) as e:
                client.get(url, follow=True)
            explain_result = e

        ctx["explain"] = explain_result
        if explain_result and explain_result.queries:
            cache.set(
                get_cache_key(explain_result.id),
                explain_result,
                timeout=getattr(settings, "DJANGO_PEV_CACHE_TIMEOUT", 60 * 60 * 24),
            )
            ctx["slowest"] = explain_result.slowest

        return ctx


class AnalyzeForm(forms.Form):
    explainset_id = forms.CharField()
    query_index = forms.CharField()


class ExplainVisualize(FormView, BaseView):
    template_name = "django_pev/pev_embedded.html"
    form_class = AnalyzeForm

    def form_valid(self, form: AnalyzeForm):
        explain_set = cache.get(get_cache_key(form.cleaned_data["explainset_id"]), None)
        if not explain_set:
            return HttpResponseRedirect(reverse("django_pev:explain"))

        assert isinstance(explain_set, ExplainSet)

        query = explain_set.queries[int(form.cleaned_data["query_index"])]
        explain_plan = query.explain()
        explain_id = str(uuid.uuid4())

        context = {
            "explain_id": explain_id,
            "explain_set": explain_set,
            "plan": explain_plan,
            "query": query,
        }

        cache.set(
            get_cache_key(explain_id),
            context,
            timeout=getattr(settings, "DJANGO_PEV_CACHE_TIMEOUT", 60 * 60 * 24),
        )

        return HttpResponseRedirect(f'{reverse("django_pev:embedded-pev")}?explain_id={explain_id}')


class EmbeddedPev(BaseView):
    template_name = "django_pev/pev_embedded.html"

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        if not self.request.GET.get("explain_id"):
            return HttpResponseRedirect(f'{reverse("django_pev:explain")}')

        explain_id = self.request.GET["explain_id"]

        explain_plan = cache.get(get_cache_key(explain_id), None)
        if not explain_plan:
            return HttpResponseRedirect(f'{reverse("django_pev:explain")}')

        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        ctx = super().get_context_data(**kwargs)
        explain_context: dict = cache.get(get_cache_key(self.request.GET["explain_id"]))
        ctx["query"] = explain_context["query"].sql
        ctx["plan"] = explain_context["plan"]
        ctx["url"] = explain_context["explain_set"].url
        ctx["duration"] = explain_context["query"].duration

        return ctx
